/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data;

import java.io.IOException;
import java.util.*;
import com.niggle.util.*;
import com.niggle.data.metadata.*;
import java.rmi.server.UnicastRemoteObject;
import org.xml.sax.SAXException;
import java.rmi.RemoteException;

/**
 *  This object plays a central role in the niggle framework. It is essentially
 * a repository of "metadata" that allows us to get at the various containers
 * as well as "exemplars" of the various data records.
 * @author <a href="mailto:jrevusky@bigfoot.com>Jonathan Revusky</a>
  */
public class DataRegistryImpl extends UnicastRemoteObject implements DataRegistry
{
    private static DataRegistry defaultRegistry;
    private static String metadataLocation;
    private HashMap dataSourceLookup = new HashMap();
    private HashMap metadataLookup = new HashMap();
    private DataRegistry remoteRegistry;

    public DataRegistryImpl() throws IOException {}

    public static DataRegistry getDefaultRegistry()
    {
        if (defaultRegistry == null)
        {
            try {
                defaultRegistry = new DataRegistryImpl();
            }
            catch (IOException ioe) {
                throw new AssertionFailedException(ioe);
            }
        }
        return defaultRegistry;
    }

    /**
     * Optional hook to call if you want to substitute a custom DataRegistry
     * implementation for the default one.
     */
    public static void setDefaultRegistry(DataRegistry registry)
    {
        defaultRegistry = registry;
    }

    /**
     * Set a second DataRegistry source that will be used if the
     * type is not found locally. This will likely be on a remote
     * machine, hence the nomenclature.
     */
    public void setRemoteRegistry(DataRegistry registry) throws IOException
    {
        remoteRegistry = registry;
    }

    public RecordDescriptor getMetadata(String type) throws IOException
    {
        RecordDescriptor result = (RecordDescriptor) metadataLookup.get(type);
        if (result == null && remoteRegistry != null)
           result = remoteRegistry.getMetadata(type);
        return result;
    }


    public RecordDescriptor getMetadata(String type, int version) throws IOException
    {
        String typeName = type + "." + version;
        RecordDescriptor result = (RecordDescriptor) metadataLookup.get(typeName);
        if (result == null && remoteRegistry != null)
        {
            result = remoteRegistry.getMetadata(type, version);
            if (result != null)
                registerRecordType(result);
        }
        return result;
    }

    public void registerRecordType(RecordDescriptor descriptor) throws IOException
    {
        String type = descriptor.getType();
        int version = descriptor.getVersion();
        String fullyQualified = type + "." + version;
        metadataLookup.put(fullyQualified, descriptor);
        // Now check if the default record is of an inferior version to this one.
        RecordDescriptor desc = (RecordDescriptor) metadataLookup.get(type);
        if (desc == null || desc.getVersion() < version)
        {
            metadataLookup.put(type, descriptor);
        }
    }

    public void registerDataSource(DataSource source) throws IOException
    {
        dataSourceLookup.put(source.getName(), source);
    }

    public void registerDataSource(DataSourceMetadata dsm) throws IOException
    {
        String name = dsm.getName();
        if (dsm.preloads())
            dataSourceLookup.put(name, dsm.getInstance());
        else
            dataSourceLookup.put(dsm.getName(), dsm);
    }


    public DataSource getDataSource(String name) throws IOException
    {
        Object value = dataSourceLookup.get(name);
        if (value instanceof DataSource)
        {
            return (DataSource) value;
        }
        else if (value instanceof DataSourceMetadata)
        {
            return ((DataSourceMetadata) value).getInstance();
        }
        if (remoteRegistry != null)
        {
            return remoteRegistry.getDataSource(name);
        }
        return null;
    }

    public Record getExemplar(String type, int version) throws IOException
    {
        RecordDescriptor desc = getMetadata(type, version);
        if (desc == null)
           return getExemplar(type);
        return desc.newRecord();
    }

    public Record getExemplar(String type) throws IOException
    {
        RecordDescriptor desc = (RecordDescriptor) getMetadata(type);
        if (desc == null)
           return null;
        return desc.newRecord();
    }

    public List getDataSourceNames() throws IOException
    {
        List result = new ArrayList(dataSourceLookup.keySet());
        if (remoteRegistry != null)
        {
            List rl = remoteRegistry.getDataSourceNames();
            for (Iterator it = rl.iterator(); it.hasNext();)
            {
                Object next = it.next();
                if (!result.contains(next))
                {
                    result.add(next);
                }
            }
        }
        return result;
    }

    public List getRecordNames() throws IOException
    {
        List result = new ArrayList(metadataLookup.keySet());
        if (remoteRegistry != null)
        {
            List rl = remoteRegistry.getRecordNames();
            for (Iterator it = rl.iterator(); it.hasNext();)
            {
                Object next = it.next();
                if (!result.contains(next))
                {
                    result.add(next);
                }
            }
        }
        return result;
    }
}
