/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data;

import java.io.IOException;
import java.util.List;
import java.util.Properties;

/**
 * a marker interface representing objects that contain
 * a set of Records indexed by a lookup key
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

public interface DataSource extends java.rmi.Remote
{
    /**
     * @return a unique string by which this DataSource will be known.
     */
    String getName() throws IOException;

    /**
     * method that exists for backward compatibility
     * This is equivalent to get(null, key);
     */
    Record get(Object key) throws IOException;

    /**
     * @param type the type of the record, if this is null,
     * then any type will do.
     * @param key the lookup key
     * @return a Record of the given type, with
     * the given lookup key. If the data source allows multiple records,
     * only returns the first one found. Use getRecords() in that situation.
     * returns null if none is found.
     */

    Record get(String type, Object key) throws IOException;

    /**
     *  @return a List of records based on a lookup key.  The result can
     *  be an empty list, but never null.
     */
    List getRecords(Object key) throws IOException;

    /**
     * This is equivalent to calling keys(null)
     * where no record type is specified.
     */
    List keys() throws IOException;

    /**
     * @param type the record type we are interested,
     * under some circumstances, this may be null.
     * @return a list of all valid lookup keys that correspond
     * to a given record type.
     */
    List keys(String type) throws IOException;


    /**
     * A method that initializes a DataSource with a set of
     * initialization properties. This is called internally
     * within niggle and should not be called by application
     * code. As an implementor, you should make sure
     * that this method is only called once, for example,
     * throw a RuntimeException if an attempt is made to call this
     * on a DataSource a second time.
     */
    void init(Properties props) throws IOException;

    /**
     *  A method that should be called to free any resources associated
     *  with the data source.  Note that no other methods should be called
     *  on this object after this one.
     *  (In many implementations, this method will likely do nothing.)
     */
    void close() throws IOException;
}
