/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data;

import java.io.*;
import java.util.*;
import com.niggle.util.*;
import javax.servlet.*;

/**
 * a holder for various useful static routines dealing
 * with niggle data
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

abstract public class DataUtil
{
    /**
     * Given a ServletRequest and a record, attempts to fill in
     * the record's fields.
     * requires that the record is mutable
     *
     * Generic heuristics for filling in a data record based on key=value
     * pairs in a ServletRequest object. In practice, this method
     * may not be enough and you may need to write some custom code
     * to fill in your records from a servlet request. Nonetheless,
     * this method provides a good starting point.
     *
     * @param record the record to fill in.
     * @param request the servlet request that provides the param=value pairs.
     */
    static public void fillRecordFromServletRequest(Record record, ServletRequest request)
    throws IOException
    {
        Assert.assert(!record.isImmutable());
        final RecordDescriptor FIELDS = record.getMetadata();
        final HashMap table = new HashMap();
        for ( Enumeration e = request.getParameterNames(); e.hasMoreElements(); )
        {
            String paramName = (String) e.nextElement();
            String paramValue = request.getParameter(paramName);
            FieldDescriptor field = FIELDS.getField(paramName);
            if (field != null)
            {
                table.put(paramName, field.normalize(paramValue));
            }
        }
        // check for required boolean fields. A bit kludgy, take into account that
        // that a checkbox will not return a param of any sort if it is unchecked.
        // Be careful, since, if there is no mention of a required boolean field
        // in the HTTP request, it gets set to false. (This is a major A-1-A gotcha.)
        for (Enumeration fields = FIELDS.getFieldDescriptors(); fields.hasMoreElements();)
        {
            FieldDescriptor field = (FieldDescriptor) fields.nextElement();
            if (field.isRequired() && field.getJavaType() == Boolean.class)
            {
                String param = request.getParameter(field.getName());
                if (!Util.hasContent(param))
                {
                    table.put(field.getName(), "false");
                }
            }
        }
        ((DefaultRecord) record).setValuesFromTable(table);
    }


    /**
     * convenience routine for transitioning records.
     * Copies all the fields in source to dest,
     * where dest has a field of the same name.
     */
    static public void copyFieldsByName(Record source, Record dest)
    {
        Assert.assert(!dest.isImmutable());
        RecordDescriptor sourceFields = source.getMetadata();
        RecordDescriptor destFields = dest.getMetadata();
        for (Enumeration e = sourceFields.getFieldDescriptors(); e.hasMoreElements();)
        {
            FieldDescriptor sourceField = (FieldDescriptor) e.nextElement();
            FieldDescriptor destField = (FieldDescriptor) destFields.getField(sourceField.getName());
            if (destField != null)
            {
                try
                {
                    Object val = source.get(sourceField);
                    dest.set(destField, val);
                } catch (InvalidFieldException ife)
                {
                    // if we can't do it, we can't do it.
                }
            }
        }
    }

    /**
     * Read list of key-value pairs representing one record in the object stream.
     * List is terminated by a blank line or EOF.
     * Ignores #comment lines.
     * @param rec the Record involved. We need this to get at the RecordDescriptor
     * metadata object.
     * @param in     Stream from which to read the record
     * @return a table of key-value pairs from the input stream, i.e. a HashTable
     * @exception IOException
     */
    static Map getFieldValueTable(Record rec, DataInput in)
    throws IOException
    {
        final HashMap result = new HashMap();
        String line = null;
        do
        {
            line = in.readLine();
            if ( line == null ) break;
            line = line.trim();
        }
        while ( line.length() == 0  || line.charAt(0) == '#' );

        // okay, now we should have got the first line.

        boolean firstLine = true; // transitional kludge

        while ( line != null && line.length() >0 )
        {
            if ( line.charAt(0) != '#' )
            {
                int index = line.indexOf('=');
                if ( index < 0 )
                    throw new MangledDataException(rec, "invalid format, no equals sign delimiter");
                String key = line.substring(0, index).toLowerCase().intern();
                if (firstLine && key == "type") // effectively change "type=" to "_meta="
                {
                    key = Record.TYPE_KEY;
                }
                String value = line.substring(++index);
                result.put(key, value);
            }
            line = in.readLine();
            if ( line != null )
                line = line.trim();
            firstLine = false;
        }
        return result;
    }

    /**
     * convenience routine to convert a string to a Record.
     * Is essentially the reverse operation of: DefaultRecord.toString()
     */
    public static Record fromString(String s)
    throws IOException, ClassNotFoundException
    {
        ByteArrayInputStream bais = new ByteArrayInputStream(s.getBytes());
        ObjectInput oi = new DefaultRecordInputStream(bais);
        return (Record) oi.readObject();
    }
}
