/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data;

import java.io.IOException;
import java.util.List;

/**
 * Interface implemented by objects that manage a collection
 * of records.
 *
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */
public interface MutableDataSource extends DataSource, DataChangeNotifier
{
    /**
     * Adds a new record to the managed pool.
     *
     * @param rec the record to add
     *
     * @exception DuplicateRecordException if another record already exists
     * with same primary key as 'rec'.
     *
     * @exception IOException if the record cannot be
     * initialized (i.e. has missing fields or invalid
     * field values, etc.), or in case of a low-level error.
     */
    void insert(Record rec) throws IOException;

    /**
     * Replaces an existing version of a record with a new updated version.
     * Automatically replaces the previous version <b>if</b> it has not been
     * modified since the object was taken from the data source.  A
     * ModifiedRecordException is thrown otherwise.
     * <p>
     * Not all underlying storage mediums will be able to support these
     * semantics.  Check your implementation carefully.
     *
     * @param rec the new updated record.
     *
     * @exception ModifiedRecordException thrown if the record was modified
     * by another process since rec was requested from a data source.
     *
     * @exception IOException thrown in case of any other database or
     * communication error.
     */
    void update(Record rec) throws IOException;

    /**
     *  Replace an existing version of a record with a new updated version.
     *  Automatically replaces the prevous version, even if it has been
     *  modified by another process.
     *  <p>
     *  The client's version of the original record is passed in order
     *  to allow implementations to make the update as fine-grained as
     *  possible.  Some will simply update the entire records, while others
     *  will update only modified fields.
     *
     *  @param rec the new updated record.
     *
     *  @param oldRec the client's copy of the record it removed from
     *  the data source.
     *
     *  @exception IOException thrown in case of a database or communication
     *  error.  Note that ModifiedRecordException is <b>never</b> thrown
     *  by this method.
     */
    void forceUpdate(Record rec, Record oldRec) throws IOException;

    /**
     * This is equivalent to delete(null, key)
     */

    void delete(Object key) throws IOException;

    /**
     *  Delete a record given its primary lookup key.
     *
     *  @param type the record type we want to delete
     *  If this is null, we delete indistinctly with no sanity check.
     *  @param key the key to the object to delete.
     *
     *  @exception MissingRecordException thrown if the record to be
     *  deleted could not be found.
     *
     *  @exception DataException thrown if the key maps to more than
     *  one record.
     *
     *  @exception IOException thrown in case of any other database or
     *  communication error.
     */
    void delete(String type, Object key) throws IOException;

    /**
     * This is equivalent to select(null, filter)
     */
    List select(RecordFilter filter) throws IOException;


    /**
     *  Fetches a list of records matching filter.  If filter is null,
     *  returns all the records in the container.
     *
     *  @param the type of record we are interested in, possibly null
     *  @param filter the record filter, or null.
     *
     *  @return a List of all the records that are an instance of a
     *  given class.  If no records match the filter, an empty List.
     */
    List select(String type, RecordFilter filter) throws IOException;

        /**
     * If this data source is backed by some external mechanism,
     * like an RDBMS, wipes the in-memory cache, if one is being
     * maintained.
     */

    void wipeCache() throws IOException;
}
