/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data;

import java.lang.ref.*;
import java.util.Locale;
import java.util.Date;
import com.niggle.util.*;

/**
 * An object that represents a weak or soft reference
 * to an underlying Record object. This object
 * implements the Record interface by delegating
 * calls to an underlying record.
 * Most likely this class will be subclassed, since
 * it will be necessary to implement the restoreReferent() method.
 * However, this could still be useful on its own, since if
 * the underlying record gets gc'ed, then the calls to the
 * various get/set methods simply "page-fault" and the
 * client code could simply catch the condition
 * and recover. In any case, the presence of this base
 * class should avoid the need to repeat the same delegation
 * bubble-gum code all over the place.
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

public class RecordReference implements Record
{
   protected Reference ref;

   //Typically, you will only wrap
   // a record that has been initialized anyway.
   private boolean initialized;
   // The following data are cached in the reference
   // wrapper, since they may be needed to
   // restore the referent.
   private String type;
   private int version;
   private Object primaryKey;

   public RecordReference(Record underlying)
   {
       this(underlying, null);
   }

   public RecordReference(Record underlying, ReferenceQueue queue)
   {
       updateCacheInfo(underlying);
       ref = createReference(underlying, queue);
   }

   public Reference createReference(Record rec, ReferenceQueue queue)
   {
       return new SoftReference(rec, queue);
   }

   public Record getUnderlyingRecord()
   {
       Record result = (Record) ref.get();
       if (result == null)
       {
           result = restoreReferent(type, version, primaryKey);
           updateCacheInfo(result);
       }
       return result;
   }

   /**
    * A method that should restore the referent (i.e. the underlying
    * record) from a persistent store such as an RDBMS.
    * The base RecordReference has no specific knowledge
    * of how to restore the referent, so it just throws
    * UnsupportedOperationException. Most likely, to make use of
    * this class, you will want to subclass it and implement
    * this method in some useful way for your application.
    * @return the underlying referenced record.
    */
   public Record restoreReferent(String type, int version, Object primaryKey)
   {
       throw new UnsupportedOperationException("Don't know how to restore referenced record.");
   }

   /**
    * updates the information held in the reference wrapper.
    */
   protected void updateCacheInfo(Record rec)
   {
       this.type = rec.getType();
       this.version = rec.getVersion();
       this.primaryKey = rec.getPrimaryKey();
       this.initialized = rec.isImmutable();
   }

   public Object get(String fieldName)
   {
       return getUnderlyingRecord().get(fieldName);
   }

   public Object get(int i)
   {
       return getUnderlyingRecord().get(i);
   }

   public Object get(FieldDescriptor field)
   {
       return getUnderlyingRecord().get(field);
   }

   public void set(String fieldName, Object o)
   {
       getUnderlyingRecord().set(fieldName, o);
   }

   public void set(int i, Object o)
   {
       getUnderlyingRecord().set(i, o);
   }

   public void set(FieldDescriptor field, Object o)
   {
       getUnderlyingRecord().set(field, o);
   }

   public void clearFields()
   {
       getUnderlyingRecord().clearFields();
   }

   public Locale getDisplayLocale()
   {
       return getUnderlyingRecord().getDisplayLocale();
   }

   public RecordDescriptor getMetadata()
   {
       return getUnderlyingRecord().getMetadata();
   }

   public Record getMutableCopy()
   {
       return getUnderlyingRecord().getMutableCopy();
   }

   public boolean isStale()
   {
       return getUnderlyingRecord().isStale();
   }

   public Date getCreationDate()
   {
       return getUnderlyingRecord().getCreationDate();
   }

   public Date getLastModified()
   {
       return getUnderlyingRecord().getLastModified();
   }

   public Object getPrimaryKey()
   {
       return primaryKey;
   }

   public void setPrimaryKey(Object o)
   {
       throw new UnsupportedOperationException();
   }

   public int getVersion()
   {
       return version;
   }

   public boolean isImmutable()
   {
       return initialized;
   }

   public String getType()
   {
       return type;
   }

   public void touch()
   {
       getUnderlyingRecord().touch();
   }

   public String toString()
   {
       return getUnderlyingRecord().toString();
   }

   public boolean equals(Object o)
   {
       if (o.getClass() != this.getClass())
           return false;
       RecordReference other = (RecordReference) o;
       return getUnderlyingRecord().equals(other.getUnderlyingRecord());
   }

   public void setMetadata(RecordDescriptor desc)
   {
       throw new UnsupportedOperationException();
   }
}
