/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data.inmemory;

import com.niggle.data.*;
import com.niggle.util.*;
import java.io.*;
import java.util.*;

/**
 *  A simple implementation of the com.niggle.data.MutableDataSource
 *  interface. All records are kept in an flat array all in RAM.
 *  If you use this container, you handle persistence by
 *  registering a DataListener via addDataListener
 *  that keeps the persistent representation in synch
 *  as records are inserted, deleted, or updated.
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */
public class InMemoryRecordSet extends AbstractMutableDataSource
{
    /**
     *  The underlying container in which we store the records.
     */
    private Map container = new HashMap();

    /**
     *  An integer to keep track of what the next integer to auto-assign to
     *  a record is.
     */
    private int nextIntegerKey;

    public InMemoryRecordSet() throws IOException
    {
        super();
    }

    /**
     * Slurp the records into this InMemoryRecordSet object with a
     * stream-based idiom. Subclasses of DataException will be
     * caught and logged. Core Java IOExceptions will bubble up.
     * NB: This method is meant to be called at initialization.
     * The loading of the records does not go through the DataListener
     * mechanism.
     */
    public void loadRecords(ObjectInput input) throws IOException
    {
        Record rec = new DefaultRecord(); // just so rec !=null
        while (rec != null)
        {
            try {
                rec = (Record) input.readObject();
                if (rec != null)
                    load(rec);
            }
            catch (DataException mde) {
                Logger.log(mde.getMessage(), Logger.LOG_WARN);
            }
            catch (ClassNotFoundException cnfe) {
                Logger.log(cnfe.getMessage(), Logger.LOG_WARN);
            }
        }
    }

    public synchronized void update(Record record) throws IOException
    {
        Assert.assert(!record.isImmutable(),
                      "Expecting a mutable record.");
        Assert.assert(!record.isStale(),
                      "This record has been deleted. " +
                      "You cannot add it to the data source.");
        record.touch();
        // The record is initialized, and hence, validated and immutable.
        freeze(record);
        Object primaryKey = record.getPrimaryKey();
        Record oldRec = (Record) container.get(primaryKey);
        if ( oldRec != null )
            markStale(oldRec);
        else
        {
            Logger.log("called update but there was no extant " +
                       "record to update", Logger.LOG_WARN);
        }
        load(record);
        fireDataEvent(new DataEvent(this,
                                   DataEvent.RECORD_UPDATED,
                                   record,
                                   oldRec));
    }

    /**
     *  Updates the entire record, regardless of whether or not
     *  it was modified by another user since it was retrieved
     *  from the data store.
     */
    public void forceUpdate(Record rec, Record oldRec) throws IOException
    {
        // TODO ... implement with the proper semantics.
        update(rec);
    }

    public synchronized void insert(Record record)
        throws IOException
    {
        Object oldKey = record.getPrimaryKey();
        Object key = oldKey;
        if (key == null || "".equals(key))
           key = nextKey();
        if (container.get(key) != null)
           throw new DuplicateRecordException(record,
                                              "The primary key "
                                              + key
                                              + " is already used.");
        record.setPrimaryKey(key);
        try {
            freeze(record);
        }
        catch (IOException ioe) {
            record.setPrimaryKey(oldKey);
            throw ioe;
        }
        load(record);
        fireDataEvent(new DataEvent(
                                   this,
                                   DataEvent.RECORD_INSERTED,
                                   record,
                                   null));
    }

    public synchronized List select(String type, RecordFilter filter) throws IOException
    {
        ArrayList result = new ArrayList();
        if (filter == null)
           filter = RecordFilter.NULL_FILTER;
        Iterator it = container.values().iterator();
        while (it.hasNext())
        {
            Record rec = (Record) it.next();
            if (rec == null)
               continue;
            if (Util.hasContent(type) && !type.equals(rec.getType()))
               continue;
            if (!filter.accept(rec))
               continue;
            result.add(rec);
        }
        result.trimToSize();
        return result;
    }

    /**
     *  Called by <code>loadRecords()</code>.
     *  <p>
     *  Unlike <code>insert()</code>, this method does not
     *  notify any listeners.
     *  <p>
     *  @param record the record to be added.
     */
    void load(Record record) throws IOException
    {
        freeze(record); // make sure the record is immutable after this point.
        Object key = record.getPrimaryKey();
        Record oldRec = (Record) container.get(key);
        if ( oldRec != record )
        {
            if (oldRec != null)
            {
                markStale(oldRec);
                container.remove(key);
            }
            if (!(record instanceof AntiRecord))
            {
                container.put(key, record);
                if (key instanceof Integer)
                {
                    int id = ((Integer) key).intValue();
                    if (id >= nextIntegerKey)
                       nextIntegerKey = id + 1;
                }
            }
        }
    }

    public synchronized Record get(String type, Object key) throws IOException
    {
        Record rec = (Record) container.get(key);
        if (rec == null && !(key instanceof Integer))
        {
            String normkey = normalizeKey(type, key);
            if (!key.equals(normkey))
               rec = (Record) container.get(normkey);
            if (rec == null) try {
               // kludge to take into account the possibility
               // that an integer key was passed in as a string.
               rec = (Record) container.get(new Integer(normkey));
            }
            catch (Exception e) {}
        }
        if (type != null
           && rec!= null
           && !type.equals(rec.getType()))
        {
           rec = null;
        }
        return rec;
    }

    public void init(Properties props) throws IOException
    {
        super.init(props);
        String filename = props.getProperty("STORE");
        if (filename.startsWith("file:/"))
            filename = filename.substring("file:/".length());
        if (new File(filename).exists())
            loadRecordsFromFlatFile(filename);
        String pfs = props.getProperty("PERSIST_FREQUENCY");
        int persistFrequency = 1000;
        if (Util.hasContent(pfs))
        {
            try {
                persistFrequency = Integer.parseInt(pfs);
            }
            catch (NumberFormatException nfe) {
                Logger.log("expecting number here, got: PERSIST_FREQUENCY=\"" + pfs + "\"");
            }
        }
        new CrudeDataPersister(this, filename, persistFrequency);
    }

    public synchronized List keys(String type) throws IOException
    {
        Set allKeys = container.keySet();
        if (!Util.hasContent(type))
           return new ArrayList(allKeys);
        ArrayList result = new ArrayList();
        for (Iterator it = allKeys.iterator(); it.hasNext();)
        {
            Object key = it.next();
            Record rec = (Record) container.get(key);
            if (rec != null && rec.getType().equals(type))
               result.add(key);
        }
        result.trimToSize();
        return result;
    }

    public synchronized void delete(String type, Object key) throws IOException
    {
        Record record = null;
        if (key instanceof Record)
            record = (Record) key;
        else
            record = get(key);
        if (type != null && record != null && !type.equals(record.getType()))
           record = null;
        if (record != null)
        {
            markStale(record);
            container.remove(record.getPrimaryKey());
            fireDataEvent(new DataEvent(this,
                                        DataEvent.RECORD_DELETED,
                                        null,
                                        record));
        }
        else
        {
            throw new MissingRecordException("No record found with key: " +
                                             key.toString() + ".");
        }
    }

    public void close() throws IOException {}

    /**
     * Returns the next available primary key,
     * assuming that it is an integer.
     */
    protected Object nextKey()
    {
        return new Integer(nextIntegerKey++);
    }

    Map getContainer()
    {
        return container;
    }

    /**
     * Wrapper around loadRecords() to read records from a flat file.
     * Creates a DefaultRecordInputStream and invokes
     * the record manager's loadRecords() method.
     * Will try to pick it up from the .### temp file if that
     * is available and filename is not.
     * NB: This method is meant to be called at initialization.
     * The loading of the records does not go through the DataListener
     * mechanism.
     */
    protected void loadRecordsFromFlatFile(String filename) throws IOException
    {
        ObjectInput input = null;
        FileLock lock = null;
        File file = new File(filename);
        if (!file.exists())
        {
            File tempFile = new File(filename + ".###");
            if ( !tempFile.exists() )
            {
                // Don't want to make this an error in case we're starting
                // with a blank file.  An error would cause the watchdog
                // to terminate the application.
                Logger.log("File " +
                           filename + " not found or accessible.",
                           Logger.LOG_WARN);
                return;
            }
            else
            {
                // If there is no file of this name, but there
                // is a filename.### we read from that.
                file = tempFile;
                Logger.log("File "
                           + filename
                           + " not found, using "
                           + tempFile,
                           Logger.LOG_WARN);
            }
        }

        try
        {
            lock = new FileLock(filename);
            input = new DefaultRecordInputStream(file);
            loadRecords(new DefaultRecordInputStream(file));
        }
        finally
        {
            if (lock != null)
               lock.dispose();
            if (input!=null)
               input.close();
        }
    }
}
