/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data.inmemory;

import com.niggle.data.*;
import com.niggle.util.*;
import java.io.*;
import java.util.*;

/**
 * An implementation of com.niggle.data.SessionManager built on top
 * of our MemoryMutableDataSource
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

public class SimpleSessionManager
    extends InMemoryRecordSet implements SessionManager
{
    public SimpleSessionManager() throws IOException
    {
    }

    /**
     *  Checks the validity of a session.
     *
     *  @param sessionID the ID of the session.
     *  @param username the ID of the user to whom the session belongs.
     *  @return <code>true</code> if the session is valid,
     *  <code>false</code> otherwise.
     */
    public boolean isValidSession(String sessionID, String username)
    {
        if (!Util.hasContent(username))
            return false;
        try
        {
            Session session = (Session) get(sessionID);
            return session.isValidSession(username);
        }
        catch ( Exception e) {
            return false;
        }
    }

    /**
     *  Creates and returns a new session.
     *
     *  @param username the ID of the user requesting the session.
     *  <br><b>require:</b> <code>(username != null) &&
     *  (username.length() > 0)</code>
     *
     *  @return the new session.
     *  <br><b>ensure:</b> <code>result != null</code>
     */
    public Session newSession(String username)
    throws IOException
    {
        SessionImpl result = (SessionImpl) DataRegistryImpl.getDefaultRegistry().getExemplar("session.1");
        result.set("username", username);
        insert(result);
        return result;
    }

    /**
     *  Invalidates the current session (the result of a logout).
     *
     *  @param sessionID the ID of the session.
     *  @return success or failure of the operation.
     */
    public boolean invalidateSession(String sessionID)
    {
        try {
            Session session = (Session) get(sessionID);
            delete(sessionID);
        }
        catch (Exception e) {
            return false;
        }
        return true;
    }

    public void touchSession(String sessionID)
    {
        try {
            SessionImpl session = (SessionImpl) get(sessionID);
            if (session == null)
                return;
            long lastModifiedTime = session.getLastModified().getTime();
            if (System.currentTimeMillis() > lastModifiedTime + 60000L)
            {
                session = (SessionImpl) session.getMutableCopy();
                session.set("last_modified", new Date());
                update(session);
            }
        }
        catch (Exception e) {
            Logger.logException(e, Logger.LOG_WARN);
        }
    }

    protected Object nextKey()
    {
        String result = null;
        while (result == null)
        {
            int r = Math.abs(rnd.nextInt());
            String possibleKey = Integer.toString(r, 36);
            if (getContainer().get(possibleKey) == null)
            {
                result = possibleKey;
            }
        }
        return result;
    }
    private Random rnd = new Random();
}
