/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data.jdbc;

import java.io.*;
import java.sql.SQLException;
import java.util.*;
import com.niggle.data.*;
import com.niggle.util.*;

/**
 * A class that is essentially a facade in front of an external
 * relational database table
 */

public class JDBCBackedRecordSet extends AbstractMutableDataSource
implements DataListener
{
    String driverClassName, jdbcURL, defaultRecordType;

    // The maximum size of the cache.
    int maxCacheSize;

    public JDBCBackedRecordSet() throws IOException
    {
    }

    /**
     * invoked when the RecordManager whose events
     * we are listening to has just deleted a record.
     */
    public void recordDeleted(DataEvent event) throws IOException
    {
        try {
            getConnection().performDeletion(event.oldRecord);
            uncacheRec(event.oldRecord);
        }
        catch (Exception e) {
            Logger.logException(e);
        }
    }

    /**
     * invoked when the RecordManager whose events
     * we are listening to has just inserted a new record.
     */
    public void recordInserted(DataEvent event) throws IOException
    {
        insert(event.newRecord);
    }

    /**
     * invoked when the RecordManager whose events
     * we are listening to has just updated an existing record.
     */
    public void recordUpdated(DataEvent event) throws IOException
    {
        update(event.newRecord);
    }

    public void init(Properties props) throws IOException
    {
        super.init(props);
        this.driverClassName = props.getProperty("JDBC_DRIVER_CLASS");
        this.jdbcURL = props.getProperty("JDBC_URL");
        this.defaultRecordType = props.getProperty("DEFAULT_RECORD_TYPE");
        String cacheSizeString = props.getProperty("MAX_CACHE_SIZE");
        try {
            Class.forName(driverClassName);
            if (Util.hasContent(cacheSizeString))
               maxCacheSize = Integer.parseInt(cacheSizeString);
        }
        catch (Exception e) {
            Assert.rethrow(e);
        }
    }

    synchronized public void delete(String type, Object key) throws IOException
    {
        if (type == null)
           type = defaultRecordType;
        int count = 0;
        Record rec = null;
        if (key instanceof Record)
            rec = (Record) key;
        else
            rec = get(key);
        if (rec == null)
        {
            throw new MissingRecordException("No record with primary key= " + key);
        }
        try {
            getConnection().performDeletion(rec);
        }
        catch (SQLException se) {
            throw new DataException(rec, se);
        }
        markStale(rec);
        uncacheRec(rec);
        fireDataEvent(new DataEvent(this, DataEvent.RECORD_DELETED, null, rec));
    }

    synchronized public void insert(Record rec) throws IOException
    {
        Object key = rec.getPrimaryKey();
        if ("".equals(key))
            key = null;
        if (key == null)
        {
            rec.setPrimaryKey(uniqueKey(rec));
        }
        freeze(rec);
        try {
            getConnection().performInsertion(rec);
        }
        catch (SQLException se) {
            throw new DataException(rec, se);
        }
        cacheRec(rec);
        DataEvent de = new DataEvent(this, DataEvent.RECORD_INSERTED, rec, null);
        fireDataEvent(de);
    }

    synchronized public void update(Record rec) throws IOException
    {
        Assert.assert(!rec.isImmutable(),
                      "Expecting a mutable record.");
        Assert.assert(!rec.isStale(),
                      "This record is stale. " +
                      "You cannot add it to the data source.");
        rec.touch();
        // The record is initialized, and hence, validated and immutable.
        freeze(rec);
        Object primaryKey = rec.getPrimaryKey();
        Record oldRec = get(primaryKey);
        try {
            getConnection().performUpdate(rec);
        }
        catch (SQLException se) {
            throw new DataException(rec, se);
        }
        if ( oldRec != null )
        {
            markStale(oldRec);
            uncacheRec(oldRec);
        }
        else
        {
            Logger.log("called update but there was no extant " +
                       "record to update", Logger.LOG_WARN);
        }
        cacheRec(rec);
        fireDataEvent(new DataEvent(this,
                                   DataEvent.RECORD_UPDATED,
                                   rec,
                                   oldRec));

    }

    public List keys(String type) throws IOException
    {
        if (!Util.hasContent(type))
           type = defaultRecordType;
        try {
            return getConnection().getPrimaryKeys(type);
        }
        catch (SQLException se) {
            throw new DataException(null, se);
        }
    }

    public Record get(String type, Object key) throws IOException
    {
        if (type == null)
           type = defaultRecordType;
        Record result = (Record) getFromCache(type, key);
        if (result != null)
           return result;
        try {
            result = getConnection().retrieveValues(type, key);
        }
        catch (SQLException se) {
            throw new DataException(result, se);
        }
        if (result != null)
        {
            freeze(result);
            cacheRec(result);
        }
        else
        {
            String normkey = normalizeKey(type, key);
            if (!normkey.equals(key))
               return get(type, normkey);
        }
        return result;
    }

    synchronized Record getFromCache(String type, Object key)
    {
        return (Record) cache.get(type + ":" + key);
    }

    synchronized Record cacheRec(Record rec)
    {
        if (maxCacheSize <=0)
           return rec;
        if (cache.size() >= maxCacheSize)
           wipeCache();
        String lookupKey = rec.getType() + ":" + rec.getPrimaryKey();
        Record oldRec = (Record) cache.get(lookupKey);
        if (oldRec != null)
        {
            if (oldRec.equals(rec))
               return oldRec;
            else
                markStale(oldRec);
        }
        cache.put(lookupKey, rec);
        return rec;
    }

    synchronized void uncacheRec(Record rec)
    {
        cache.remove(rec.getType() + ":" + rec.getPrimaryKey());
    }

    synchronized public void wipeCache()
    {
        Map oldCache = this.cache;
        cache = new HashMap();
        for (Iterator it = oldCache.values().iterator(); it.hasNext();)
        {
            Record rec = (Record) it.next();
            markStale(rec);
        }
    }

    public List select(String type, RecordFilter filter) throws IOException
    {
        if (!Util.hasContent(type))
           type = defaultRecordType;
        List result = null;
        try {
            result = getConnection().selectRecords(type, filter);
        }
        catch (SQLException se) {
              throw new DataException(null, se);
        }
        int listSize = result.size();
        boolean cacheRecs = listSize < maxCacheSize/2; // INTERESTING PROBLEM: REVISIT (JR)
        for (int i = 0; i<listSize; i++)
        {
            Record rec = (Record) result.get(i);
            freeze(rec);
            if (cacheRecs)
            {
                result.set(i, cacheRec(rec));
            }
        }
        return result;
    }

    // TODO: Implement the possibility of a connection pool.
    // for efficiency sake (JR)
    JDBCConnection getConnection()
    {
        try {
            Assert.assert(!jdbcConnection.isClosed());
        }
        catch (Exception e) {
            jdbcConnection = new JDBCConnection(jdbcURL);
        }
        return jdbcConnection;
    }

    private Object uniqueKey(Record rec) throws IOException
    {
        FieldDescriptor fd = rec.getMetadata().getPrimaryKeyField();
        boolean isInt = (fd.getJavaType() == Integer.class);
        int nextInt = Math.abs(rnd.nextInt());
        while (true)
        {
            Object next = new Integer(nextInt);
            if (!isInt)
            {
                next = next.toString();
            }
            if (get(next) == null)
            {
                return next;
            }
        }
    }

    private Random rnd = new Random(); // used to generate keys
    private HashMap cache = new HashMap();
    private JDBCConnection jdbcConnection;
}
