/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data.jndi;

import java.io.*;
import java.util.*;
import javax.naming.*;
import javax.naming.directory.*;
import com.niggle.data.*;
import com.niggle.util.*;

abstract public class AbstractJNDIRecord extends DefaultRecord
    implements DirContext
{
    // Constants.

    private static final String OBJECTCLASS_KEY = "objectclass";

    // Attributes.

    protected Attributes attrs;

    public AbstractJNDIRecord()
    {
        attrs = new BasicAttributes(true);
        Attribute oc = new BasicAttribute(OBJECTCLASS_KEY);
        oc.add("extensibleObject");
        oc.add("top");
        attrs.put(oc);
    }

    /**
     *  Called by a JNDI data source in order to populate the attributes
     *  from the data table.
     */
    public void populateAttributes()
    {
        // Set the type.
        attrs.put(TYPE_KEY, getType() + "." + getVersion());

        Enumeration fields = getMetadata().getFieldDescriptors();
        while (fields.hasMoreElements())
        {
            FieldDescriptor field = (FieldDescriptor) fields.nextElement();
            String value = null;
            try {
                Object o = get(field);
                if (o != null)
                    value = field.valueToString(o);
            }
            catch (IOException ioe) {
                throw new AssertionFailedException("It should be " +
                                                   "impossible to get here.");
            }
            if (!Util.hasContent(value))
                continue;
            attrs.put(field.getName(), value);
        }
    }

    /**
     *  Get all the attributes associated with this context.  Note
     *  that we don't support named objects internally.
     *
     *  @param name the name of the object whose attributes we want.
     *  This must be <b>and empty String</b> since we don't support
     *  named objects internally.  By convention (?) an empty String
     *  implies the current context.
     */
    public Attributes getAttributes(String name) throws NamingException
    {
        if (!"".equals(name))
            throw new NameNotFoundException();
        return (Attributes) attrs.clone();
    }

    public Attributes getAttributes(Name name) throws NamingException
    {
        return getAttributes(name.toString());
    }

    /**
     *  Return only the desired attributes.
     */
    public Attributes getAttributes(String name, String[] ids)
        throws NamingException
    {
        Attributes result = new BasicAttributes(true);
        Attribute attr;

        if (!"".equals(name))
            throw new NameNotFoundException();

        for (int i = 0; i < ids.length; i++)
        {
            attr = attrs.get(ids[i]);
            if (attr != null)
                result.put(attr);
        }

        return result;
    }

    public Attributes getAttributes(Name name, String[] ids)
        throws NamingException
    {
        return getAttributes(name.toString(), ids);
    }

    /**
     *  Unsupported
     */
    public Object lookup(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Object lookup(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void bind(Name name, Object obj) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void bind(String name, Object obj) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void rebind(Name name, Object obj) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void rebind(String name, Object obj) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void unbind(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void unbind(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void rename(Name oldName, Name newName) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void rename(String oldName, String newName) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration list(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration list(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration listBindings(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration listBindings(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void destroySubcontext(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void destroySubcontext(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Context createSubcontext(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Context createSubcontext(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Object lookupLink(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Object lookupLink(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NameParser getNameParser(Name name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NameParser getNameParser(String name) throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public String composeName(String name, String prefix)
	    throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Name composeName(Name name, Name prefix)
	    throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Object addToEnvironment(String propName, Object propVal)
	throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Object removeFromEnvironment(String propName)
	throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public Hashtable getEnvironment() throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void close() throws NamingException {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void modifyAttributes(Name name, int mod_op, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void modifyAttributes(String name, int mod_op, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void modifyAttributes(Name name, ModificationItem[] mods)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void modifyAttributes(String name, ModificationItem[] mods)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void bind(Name name, Object obj, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void bind(String name, Object obj, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void rebind(Name name, Object obj, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public void rebind(String name, Object obj, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public DirContext createSubcontext(Name name, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public DirContext createSubcontext(String name, Attributes attrs)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public DirContext getSchema(Name name) throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public DirContext getSchema(String name) throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public DirContext getSchemaClassDefinition(Name name)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public DirContext getSchemaClassDefinition(String name)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(Name name,
				    Attributes matchingAttributes,
				    String[] attributesToReturn)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(String name,
				    Attributes matchingAttributes,
				    String[] attributesToReturn)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(Name name,
				    Attributes matchingAttributes)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(String name,
				    Attributes matchingAttributes)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(Name name,
				    String filter,
				    SearchControls cons)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(String name,
				    String filter,
				    SearchControls cons)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(Name name,
				    String filterExpr,
				    Object[] filterArgs,
				    SearchControls cons)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public NamingEnumeration search(String name,
				    String filterExpr,
				    Object[] filterArgs,
				    SearchControls cons)
	    throws NamingException  {
	throw new UnsupportedOperationException();
    }

    /**
     *  Unsupported
     */
    public String getNameInNamespace() throws NamingException {
	throw new UnsupportedOperationException();
    }
}
