/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*
 *  IMPLEMENTATION NOTES:
 *
 *  All _xxx() functions take a key argument expect the key to
 *  be fully qualified, i.e., any prefix or suffix should already
 *  have been added to it.
 *
 *  Only xxx() should call its _xxx() cousin.
 */


package com.niggle.data.jndi;

import java.io.*;
import java.util.*;
import javax.naming.*;
import javax.naming.directory.*;
import com.niggle.data.*;
import com.niggle.util.*;

/**
 *  A JNDI implementation of the com.niggle.data.MutableDataSource
 *  interface.
 *
 * @author <a href="mailto:murphy@niggle.com">Paul Murphy</a>
 */
public class JNDIMutableDataSource extends AbstractMutableDataSource
{
    private DirContext _ctx = null;
    private Hashtable _env = null;
    private String _indexField = null; // Will not be necessary eventually
    private String _indexFieldPrefix = null;
    private String _indexFieldSuffix = null;

    // REQUIRED PROPERTY KEYS.

    public static final String INITIAL_CONTEXT_FACTORY_KEY = "CONTEXT_FACTORY";
    public static final String PROVIDER_URL_KEY = "PROVIDER_URL";
    public static final String SEC_PRINCIPAL_KEY = "SEC_PRINCIPAL";
    public static final String SEC_CREDENTIALS_KEY = "SEC_CREDENTIALS";
    public static final String OBJECT_FACTORIES_KEY = "OBJECT_FACTORIES";
    public static final String INDEX_FIELD_KEY = "INDEX_FIELD";

    // OPTIONAL PROPERTY KEYS.

    public static final String INDEX_FIELD_PREFIX_KEY = "INDEX_FIELD_PREFIX";
    public static final String INDEX_FIELD_SUFFIX_KEY = "INDEX_FIELD_SUFFIX";

    public JNDIMutableDataSource() throws IOException {super();}

    public void init(Properties props) throws IOException
    {
        super.init(props);

        // Make sure have all the necessary data to connect to our data
        // source.

        String contextFactory = props.getProperty(INITIAL_CONTEXT_FACTORY_KEY);
        String providerURL = props.getProperty(PROVIDER_URL_KEY);
        String securityPrincipal = props.getProperty(SEC_PRINCIPAL_KEY);
        String securityCredentials = props.getProperty(SEC_CREDENTIALS_KEY);
        String objectFactories = props.getProperty(OBJECT_FACTORIES_KEY);
        _indexField = props.getProperty(INDEX_FIELD_KEY);
        _indexFieldPrefix = props.getProperty(INDEX_FIELD_PREFIX_KEY);
        _indexFieldSuffix = props.getProperty(INDEX_FIELD_SUFFIX_KEY);

        if ((contextFactory == null) ||
            (providerURL == null) ||
            (securityPrincipal == null) ||
            (securityCredentials == null) ||
            (objectFactories == null) ||
            (_indexField == null))
        {
            throw new MissingDataException("A parameter required for the " +
                                           "initialization of a " +
                                           "JNDIMutableDataSource is missing");
        }

        // If the indexFieldPrefix is null, make it an empty String.
        if (_indexFieldPrefix == null)
            _indexFieldPrefix = "";

        // If the indexFieldSuffix is null, make it an empty String.
        if (_indexFieldSuffix == null)
            _indexFieldSuffix = "";

        // Setup the environment needed to make the connection.

        _env = new Hashtable();
        _env.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
        _env.put(Context.PROVIDER_URL, providerURL);
        _env.put(Context.SECURITY_PRINCIPAL, securityPrincipal);
        _env.put(Context.SECURITY_CREDENTIALS, securityCredentials);
        _env.put(Context.OBJECT_FACTORIES, objectFactories);

        // Connect.

        connect();

        // TODO ... a separate thread should be started to see if the
        // connection is up and attempt to re-connect if it goes
        // down...see how JR implemented this in the MediMOM application
        // code.
    }

    public void close() throws IOException
    {
        if (context() != null)
        {
            try {
                context().close();
            }
            catch (NamingException e) {
                Logger.logException(e);
                throw new IOException(e.getMessage());
            }
        }
    }

    public synchronized void update(Record record) throws IOException
    {
        throw new UnsupportedOperationException();
    }

    public synchronized void forceUpdate(Record rec, Record oldRec)
        throws IOException
    {
        Assert.assert(!rec.isImmutable(), "Expecting a mutable record.");
        Assert.assert(!rec.isStale(), "This record has been deleted. " +
                      "You cannot add it to a DataSource.");
        Assert.assert(rec instanceof AbstractJNDIRecord,
                      "This record is not an instance of DirContext.");


        freeze(rec);

        if (oldRec == null)
        {
            try {
                oldRec = (Record) get(rec.get(_indexField));
            }
            catch (MissingRecordException e) {
                Logger.log("Called update but there was no extant record " +
                           "to update.", Logger.LOG_WARN);
            }
        }

        _update(rec, oldRec);

        fireDataEvent(new DataEvent(this,
                                   DataEvent.RECORD_UPDATED,
                                   rec,
                                   oldRec));
    }

    public synchronized void insert(Record rec) throws IOException
    {
        Assert.assert(!rec.isStale(), "This record has been deleted. " +
                      "You cannot add it to a DataSource.");
        Assert.assert((rec.getPrimaryKey() != null),
                      "This record's primary key has not been set.");
        Assert.assert(rec instanceof DirContext, "This record is not an " +
                      "instance of DirContext.");

        freeze(rec);

        _insert(rec);

        fireDataEvent(new DataEvent(this, DataEvent.RECORD_INSERTED, rec,
                                    null));
    }

    public synchronized void delete(String type, Object key) throws IOException
    {
        // Will throw the MissingRecordException is not there.
        Record rec = get(type, key);
        _delete(fullKey(key));
        fireDataEvent(new DataEvent(this,
                                    DataEvent.RECORD_DELETED,
                                    null,
                                    rec));
    }

    public synchronized Record get(String type, Object key) throws IOException
    {
        return _get(fullKey(key));
    }

    public synchronized List select(String type, RecordFilter filter) throws IOException
    {
        ArrayList result = new ArrayList();

        if (filter == null)
           filter = RecordFilter.NULL_FILTER;

        // TODO...deal with other filters.

        try {
            NamingEnumeration list = context().listBindings("");
            while (list.hasMore())
            {
                Record rec = (Record) ((Binding) list.next()).getObject();
                if (Util.hasContent(type) && !type.equals(rec.getType()))
                   continue;
                result.add(rec);
            }
        }
        catch (NamingException e) {
            Logger.logException(e, Logger.LOG_ERROR);
            throw new IOException(e.getMessage());
        }

        return result;
    }

    public synchronized List keys(String type) throws IOException
    {
        List result = new ArrayList();
        try {
            NamingEnumeration list = context().list("");
            while (list.hasMore())
            {
                String fullKey = ((NameClassPair) list.next()).getName();
                result.add(key(fullKey));
            }
        }
        catch (NamingException e) {
            Logger.logException(e, Logger.LOG_ERROR);
            throw new IOException(e.getMessage());
        }
        return result;
    }

    private DirContext context() {return _ctx;}

    private void connect()
    {
        try {
            _ctx = new InitialDirContext(_env);
        }
        catch (NamingException e) {
            Logger.log("Unable to retrieve CCAppUsers context.",
                       Logger.LOG_ERROR);
        }
    }

    private void _update(Record newRec, Record oldRec) throws IOException
    {
        String fullKey = fullKey(newRec.get(_indexField));

        Logger.log("Updating object with key " + fullKey, Logger.LOG_DEBUG);

        try {
            AbstractJNDIRecord _rec = (AbstractJNDIRecord) newRec;
            _rec.populateAttributes();
            // TODO ... be selective!!!
//            Attributes attrs = _rec.getAttributes("");
//            context().rebind(fullKey, newRec, attrs);
            context().rebind(fullKey, newRec);
        }
        catch (NamingException e) {
            Logger.logException(e, Logger.LOG_ERROR);
            throw new IOException(e.getMessage());
        }
    }

    private void _insert(Record rec) throws IOException
    {
        Object key = rec.get(_indexField);
        String fullKey = fullKey(key);

        Logger.log("Inserting object with key " + fullKey, Logger.LOG_DEBUG);

        // Make sure we don't already have one with that key.
        try {
            get(key);
            throw new DuplicateRecordException(key + " alrady exists.");
        }
        catch (MissingRecordException e) {
        }

        try {
            AbstractJNDIRecord _rec = (AbstractJNDIRecord) rec;
            _rec.populateAttributes();
//            Attributes attrs = _rec.getAttributes("");
//            context().bind(fullKey, rec, attrs);
            context().bind(fullKey, rec);
        }
        catch (NamingException e) {
            Logger.logException(e, Logger.LOG_ERROR);
            throw new IOException(e.getMessage());
        }
    }

    private void _delete(String fullKey) throws IOException
    {
        Logger.log("Deleting object with key " + fullKey, Logger.LOG_DEBUG);

        try {
            context().unbind(fullKey);
        }
        catch (NamingException e) {
            Logger.logException(e, Logger.LOG_ERROR);
            throw new IOException(e.getMessage());
        }
    }

    private Record _get(String fullKey) throws IOException
    {
        Record result = null;

        Logger.log("Retrieving object with key " + fullKey, Logger.LOG_DEBUG);

        try {
            Object o = context().lookup(fullKey);
            if (o instanceof Record)
                result = (Record) o;
        }
        catch (NamingException e) {
        }

        return result;
    }

    /**
     *  Constructs a full key from the prefix, suffix, and base key.
     */
    private String fullKey(Object key)
    {
        return _indexFieldPrefix + key.toString() + _indexFieldSuffix;
    }

    /**
     *  Constructs a key from a full key by eliminating the prefix and
     *  suffix.
     */
    private Object key(String fullKey)
    {
        Object result = null;

        int start = _indexFieldPrefix.length();

        if (_indexFieldSuffix.length() != 0)
        {
            int end = fullKey.length() - _indexFieldSuffix.length();
            result = fullKey.substring(start, end);
        }
        else
            result = fullKey.substring(start);

        return result;
    }
}
