/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.data.metadata;

import com.niggle.data.*;
import com.niggle.util.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import java.io.IOException;
import java.util.*;
import java.text.*;

/**
 * A descriptor for a field that holds a date or timestamp
 * (probably a misnomer, should be called TimeStampField)
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

public class DateField extends Field
{
    private boolean setToNow;
    private long defaultOffsetFromNow;

    public DateField() {}

    protected void init(Element element) throws SAXException
    {
        NodeList formats = element.getElementsByTagName("FORMAT");
        int numFormats = formats.getLength();
        if (numFormats >0)
        {
            this.dateFormat = getDateFormat((Element) formats.item(0));
            alternateDateFormats = new DateFormat[numFormats - 1];
            for (int i = 1; i<numFormats; i++)
            {
                alternateDateFormats[i-1] = getDateFormat((Element) formats.item(i));
            }
        }
        super.init(element);
        if (!Util.hasContent(sqlType))
           this.sqlType = "DATETIME";
    }

    protected void processDefaultValue() throws IOException
    {
        this.setToNow = "$now".equals(defaultValueString);
        if (!setToNow)
           super.processDefaultValue();
    }

    public Class getJavaType()
    {
        return dateClass;
    }

    public Object defaultValue()
    {
        return setToNow ? new Date() : null;
    }

    public Object getObjectFromString(Record rec, String value)
    throws MangledDataException
    {
        if (!Util.hasContent(value))
           return defaultValue();
        try {
            return dateFormat.parse(value);
        }
        catch (ParseException pe) {
            return checkAlternateDateFormats(rec, value);
        }
    }

    public Object getCopy(Object o)
    {
        if (o == null)
           return null;
        return ((Date) o).clone();
    }

    public String valueToString(Object o)
    {
        return dateFormat.format((Date) o);
    }

    Date checkAlternateDateFormats(Record rec, String value) throws MangledDataException
    {
        for (int i = 0; i<alternateDateFormats.length; i++)
        {
            try {
                return alternateDateFormats[i].parse(value);
            }
            catch (ParseException pe) {
            }
        }
        throw new MangledDataException(rec, "Field " + getName() + " must be a parseable date string.");
    }

    DateFormat getDateFormat(Element elem)
    {
        Locale locale = Locale.getDefault();
        String lang = elem.getAttribute("LANGUAGE");
        String country = Util.noNull(elem.getAttribute("COUNTRY"));
        String variant = Util.noNull(elem.getAttribute("VARIANT"));
        if (Util.hasContent(lang))
        {
            locale = new Locale(lang, country, variant);
        }
        String pattern = DOMUtil.getText(elem);
        return new SimpleDateFormat(pattern, locale);
    }

    private DateFormat dateFormat = new SimpleDateFormat("H:mm d MMM, yyyy", Locale.US);
    private DateFormat[] alternateDateFormats;
    static final Class dateClass = Date.class;
}
