/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL) 
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors 
may be used to endorse or promote products derived from this software 
without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

package com.niggle.data.metadata;

import com.niggle.data.*;
import com.niggle.util.*;
import org.w3c.dom.*;
import java.io.IOException;
import java.util.*;
import org.xml.sax.SAXException;

/**
 * A descriptor for a field that represents
 * an array of integer values
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

public class IntArrayField extends Field
{
    private boolean sorted, hasMin, hasMax;
    private int min, max;

    static final Class intArrayClass = (new int[0]).getClass();

    public IntArrayField() {}

    public Class getJavaType()
    {
        return intArrayClass;
    }

    public void init(Element element) throws SAXException
    {
        String sorted = element.getAttribute("SORTED");
        this.sorted = DOMUtil.isAffirmative(sorted);
        NodeList mins = element.getElementsByTagName("MINIMUM");
        int len = mins.getLength();
        if (len  > 0)
        {
            if (len >1)
               throw new SAXException("The MINIMUM of a numerical field can only be specified once.");
            try {
                Element elem = (Element) mins.item(0);
                String value = elem.getAttribute("VALUE");
                min = Integer.parseInt(value);
                hasMin = true;
            }
            catch (NumberFormatException nfe) {
                throw new SAXException("Expected an integer here.");
            }
        }
        NodeList maxes = element.getElementsByTagName("MAXIMUM");
        len = maxes.getLength();
        if (len  > 0)
        {
            if (len >1)
               throw new SAXException("The MAXIMUM of a numerical field can only be specified once.");
            try {
                Element elem = (Element) maxes.item(0);
                String value = elem.getAttribute("VALUE");
                max = Integer.parseInt(value);
                hasMax = true;
            }
            catch (NumberFormatException nfe) {
                throw new SAXException("Expected a number here.");
            }

        }
        if (hasMin && hasMax)
        {
            if (min >= max)
               throw new SAXException("Minimum value must be less than maximum.");
        }
        super.init(element);
    }

    public Object getObjectFromString(Record rec, String value)
    throws MangledDataException
    {
        StringTokenizer stok = new StringTokenizer(value, " ,");
        Vector v = new Vector();
        while ( stok.hasMoreTokens() )
        {
            v.addElement(stok.nextToken());
        }
        int[] result = new int[v.size()];
        for ( int i=0; i<result.length; i++ )
        {
            String s = (String) v.elementAt(i);
            try
            {
                result[i] = Integer.parseInt(s);
            }
            catch ( NumberFormatException nfe )
            {
                throw new MangledDataException(rec, "Field " + getName() + ", found " + s + ", was expecting an integer");
            }
        }
        if ( sorted )
        {
            Arrays.sort(result);
        }
        return result;
    }

    public Object getCopy(Object o)
    {
        if (o == null)
           return null;
        return ((int[]) o).clone();
    }

    public String valueToString(Object o)
    {
        int[] ints = (int[]) o;
        String result = "";
        for ( int i=0; i<ints.length; i++ )
        {
            result += ints[i];
            result += " ";
        }
        return result;
    }

    public void checkValidValue(Record rec, Object value) throws IOException
    {
        super.checkValidValue(rec, value);
        if (value == null)
           return;
        int[] arr = (int[]) value;
        if (hasMin)
           for (int i = 0; i< arr.length; i++)
           {
               if (min > arr[i])
                  throw new MangledDataException(rec, "Element " + i
                        + " is below the specified minimum of " + min);
           }
        if (hasMin)
           for (int i = 0; i< arr.length; i++)
           {
               if (max <= arr[i])
                  throw new MangledDataException(rec, "Element " + i
                        + " exceeds the specified maximum of " + max);
           }
    }
}
