/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL) 
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors 
may be used to endorse or promote products derived from this software 
without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

package com.niggle.data.metadata;

import com.niggle.data.*;
import com.niggle.util.*;
import org.w3c.dom.*;
import java.io.IOException;
import java.util.*;
import org.xml.sax.SAXException;
import java.net.*;

/**
 * A descriptor for a field that represents
 * an array of String values
 *
 * @author <a href="mailto:murphy@niggle.com">Paul Murphy</a>
 */

public class StringArrayField extends Field
{
    private boolean sorted;
    int min_elements = 0;
    int max_elements = Integer.MAX_VALUE;

    static final Class stringArrayClass = (new String[0]).getClass();

    public StringArrayField() {}

    public Class getJavaType()
    {
        return stringArrayClass;
    }

    public void init(Element element) throws SAXException
    {
        // Sort?
        String sorted = element.getAttribute("SORTED");
        this.sorted = DOMUtil.isAffirmative(sorted);

        // Min?
        String _min = element.getAttribute("MIN_ELEMENTS");
        if ((_min != null) && (!_min.equals("")))
        {
            try {
                min_elements = Integer.parseInt(_min);
            }
            catch (NumberFormatException e) {
                throw new SAXException("Expected MIN_ELEMENTS (" + _min +
                                       ") to be an integer.");
            }
        }
        
        // Max?
        String _max = element.getAttribute("MAX_ELEMENTS");
        if ((_max != null) && (!_max.equals("")))
        {
            try {
                max_elements = Integer.parseInt(_max);
            }
            catch (NumberFormatException e) {
                throw new SAXException("Expected MAX_ELEMENTS (" + _max +
                                       ") to be an integer.");
            }
        }

        super.init(element);
    }

    public Object getObjectFromString(Record rec, String value)
        throws MangledDataException
    {
        StringTokenizer stok = new StringTokenizer(value, " ");
        Vector v = new Vector();
        while (stok.hasMoreTokens())
        {
            String s = (String) stok.nextToken();
            try {
                v.addElement(URLDecoder.decode(s));
            }
            catch (Exception e) {
                Logger.log("Unable to decode \"" + s + "\".", 
                           Logger.LOG_ERROR);
            }
        }

        String[] result = new String[v.size()];
        for (int i = 0; i < result.length; i++)
            result[i] = (String) v.elementAt(i);

        if (sorted)
            Arrays.sort(result);

        return result;
    }

    public Object getCopy(Object o)
    {
        if (o == null)
           return null;
        return ((String[]) o).clone();
    }

    public String valueToString(Object o)
    {
        String[] strings = (String[]) o;
        String result = "";
        for (int i=0; i < strings.length; i++)
        {
            result += URLEncoder.encode(strings[i]);
            result +=" ";
        }
        return result;
    }

    public void checkValidValue(Record rec, Object value) throws IOException
    {

        super.checkValidValue(rec, value);

        if (value == null)
            return;

        String[] arr = (String[]) value;
        if (arr.length < min_elements)
        {
            throw new MangledDataException("Field " + getName() + " contains fewer values than allowed:  " + Integer.toString(min_elements) + ".");
        }

        if (arr.length > max_elements)
        {
            throw new MangledDataException("Field " + getName() + " contains more values than allowed:  " + Integer.toString(max_elements) + ".");
        }
    }
}
