/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.servlet;

import javax.servlet.*;
import javax.servlet.http.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;
import java.rmi.RemoteException;
import com.niggle.data.*;
import com.niggle.templates.*;
import com.niggle.util.*;
import com.niggle.data.metadata.MetadataLoader;

/**
 * A custom implementation of the ServletConfig interface
 * The default NiggleServlet initializations are delegated
 * to this object.
 * @see com.niggle.servlet.NiggleServlet#init
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

public class NiggleConfig implements ServletConfig, NiggleConstants
{
    private ServletConfig nestedConfig = NULL_CONFIG;
    private Properties params = new Properties();
    private Constructor siConstructor;
    private PageFactory pageFactory;
    private HashMap methodLookup = new HashMap();
    private static final Class[] EMPTY_CLASS_ARRAY = new Class[0];
    private ClassLoader servletClassLoader;
    static final String EXEC_PREFIX = "exec";

    public NiggleConfig() {}

    public NiggleConfig(ServletConfig nestedConfig, Servlet servlet) throws ServletException
    {
        servletClassLoader = servlet.getClass().getClassLoader();
        if (nestedConfig != null)
        {
           this.nestedConfig = nestedConfig;
           for (Enumeration en = nestedConfig.getInitParameterNames(); en.hasMoreElements();)
           {
               String key = (String) en.nextElement();
               String value = nestedConfig.getInitParameter(key);
               setInitParameter(key, value);
           }
        }
        initLogFile();
        loadMetadata();
        initPageFactory();
        initBaseClasses();
    }

    public Enumeration getInitParameterNames()
    {
        return params.propertyNames();
    }

    public String getInitParameter(String name)
    {
        return params.getProperty(name);
    }

    public ServletContext getServletContext()
    {
        return nestedConfig.getServletContext();
    }

    public void setInitParameter(String key, String value)
    {
        params.setProperty(key, value);
    }

    public String getServletName()
    {
        return Util.noNull(getInitParameter("SERVLET_NAME"));
    }

    protected void loadMetadata() throws ServletException
    {
        String recordMetadataLocation = getInitParameter("RECORD_METADATA");
        String dataSourceLocation = getInitParameter("DATA_SOURCES");
        try {
            DataRegistryImpl registry  = (DataRegistryImpl) DataRegistryImpl.getDefaultRegistry();
            if (Util.hasContent(recordMetadataLocation))
            {
                Logger.log("loading record metadata from: " + recordMetadataLocation, Logger.LOG_INFO);
                MetadataLoader.loadRecordMetadata(registry, recordMetadataLocation, servletClassLoader);
            }
            if (Util.hasContent(dataSourceLocation))
            {
                Logger.log("loading data source info from: " + dataSourceLocation, Logger.LOG_INFO);
                MetadataLoader.loadDataSources(registry, dataSourceLocation, servletClassLoader);
            }
        }
        catch (Exception e) {
            Logger.logException(e, Logger.LOG_ERROR);
            throw new ServletException(e);
        }
    }

    protected void initPageFactory() throws ServletException
    {
        String resourceBase = getInitParameter("RESOURCE_BASE");
        String pageFactoryClass = getInitParameter("PAGE_FACTORY_CLASS");
        if (!Util.hasContent(pageFactoryClass))
            pageFactoryClass = "com.niggle.templates.freemarkerimpl.FreemarkerPageFactory";
        try {
            this.pageFactory = (PageFactory) Util.findClass(pageFactoryClass, servletClassLoader).newInstance();
            if (Util.hasContent(resourceBase))
                pageFactory.setDefaultPath(resourceBase);
        }
        catch (Exception e) {
            throw new ServletException(e);
        }
        Logger.log("Initialized page factory.", Logger.LOG_INFO);
        Logger.log("Resource base is " + resourceBase + ".", Logger.LOG_INFO);
        if (!Util.hasContent(resourceBase))
        {
            Logger.log("Will load page templates relative to the servlet context classpath", Logger.LOG_INFO);
        }
    }

    protected void initLogFile() throws ServletException
    {
        String logFile = getInitParameter("LOGFILE");
        if (Util.hasContent(logFile))
        {
            Logger.setLogFileName(logFile);
        }
        Logger.log("Initializing servlet.", Logger.LOG_INFO);
    }

    protected void initBaseClasses() throws ServletException
    {
        String siClassName = getInitParameter("INTERACTION_CLASS");
        String rtClassName = getInitParameter("RECORD_TEMPLATE_CLASS");
        if (!Util.hasContent(siClassName))
            throw new ServletException("Must specify INTERACTION_CLASS, a subclass of com.niggle.ServletInteraction");
        try {
            Class siClass = Util.findClass(siClassName, servletClassLoader);
            siConstructor = siClass.getConstructor(new Class[] {
                    HttpServletRequest.class,
                    HttpServletResponse.class,
                    NiggleConfig.class
                }
            );
            if (!Util.hasContent(rtClassName))
               rtClassName = "com.niggle.templates.freemarkerimpl.RecordTemplate";
            PresentationWrapper.setRecordTemplateClass(Util.findClass(rtClassName, servletClassLoader));
        }
        catch (Exception e) {
            throw new ServletException(e);
        }
    }

    public PageFactory getPageFactory()
    {
        return this.pageFactory;
    }

    public SessionManager getSessionManager() throws IOException
    {
        SessionManager result = null;
        DataRegistry registry = DataRegistryImpl.getDefaultRegistry();
        result = (SessionManager) registry.getDataSource(SESSION_MANAGER_KEY);
        if (result == null)
        {
            String url = getInitParameter("SESSION_SERVER_URL");
            if (Util.hasContent(url)) try {
                result = (SessionManager) java.rmi.Naming.lookup(url);
                //registry.registerDataSource(result);
            }
            catch (Exception e) {
                Logger.log("Unable to contact session manager.", Logger.LOG_ERROR);
            }
        }
        return result;
    }


    /**
     * where the servlet gives us a transaction to process.
     * @param request the data about the user's request.
     * @param response the HTML response.
     * @return a ServletTransaction to handle the work.
     */
    public ServletInteraction newServletInteraction(HttpServletRequest request,
                                                    HttpServletResponse response)
    throws IOException
    {
        ServletInteraction result = null;
        try {
            result = (ServletInteraction) siConstructor.newInstance(new Object[] {
                    request, response, this
                }
            );
        }
        catch (java.lang.reflect.InvocationTargetException e) {
            Exception te = (Exception) e.getTargetException();
            if (te instanceof IOException)
               throw (IOException) te;
            else
                Assert.rethrow(te);
        }
        catch (Exception e2) {
            Assert.rethrow(e2);
        }
        return result;
    }

    /**
     * Dispatches the request based on the action
     */
    protected void dispatchAction(ServletInteraction si) throws IOException
    {
        try {
            Method method = getMethodFromAction(si, si.action);
            method.invoke(si, EMPTY_CLASS_ARRAY);
        }
        catch (IllegalAccessException iae) {
            throw new AssertionFailedException(iae);
        }
        catch (InvocationTargetException ite) {
            Throwable t = ite.getTargetException();
            if (t instanceof RemoteException)
            {
                Throwable detail = ((RemoteException) t).detail;
                if ((detail instanceof IOException) || (detail instanceof RuntimeException))
                    t = detail;
            }
            if (t instanceof IOException)
                throw (IOException) t;
            Assert.rethrow(t);
        }
        catch (NoSuchMethodException nsme) {
            Logger.logException(nsme, Logger.LOG_WARN);
            si.execDefault();
        }
    }

    protected Method getMethodFromAction(ServletInteraction si, String action)
        throws IllegalAccessException, NoSuchMethodException
    {
        String lookupName = si.getClass().getName() + "_" + action;
        if (methodLookup.containsKey(lookupName))
        {
            Method m = (Method) methodLookup.get(lookupName);

            if (m == null)
                throw new NoSuchMethodException(lookupName);
            else
                return m;
        }
        methodLookup.put(lookupName, null);
        String methodName = methodNameFromAction(action);
        Method result = si.getClass().getMethod(methodName, EMPTY_CLASS_ARRAY);
        methodLookup.put(lookupName, result);
        return result;
    }

    private String methodNameFromAction(String action)
    {
        return EXEC_PREFIX
               + Character.toUpperCase(action.charAt(0))
               + action.substring(1);
    }

    /**
     * Gets the specified input stream, assuming that it
     * may be a fully specified filename, OR relative to
     * the classpath.
     */
    public InputStream getInputStream(String location)
    {
        return Util.getInputStream(location, servletClassLoader);
    }

    // A dummy ServletConfig for the case where none is specified.

    private static ServletConfig NULL_CONFIG = new ServletConfig() {
        public Enumeration getInitParameterNames() {return Util.EMPTY_ENUM;}
        public String getInitParameter(String name) {return null;}
        public ServletContext getServletContext() {return null;}
        public String getServletName() {return "";}
    };
}
