/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.servlet;

import com.niggle.data.SessionManager;
import com.niggle.templates.Page;
import com.niggle.util.*;
import java.io.IOException;
import java.net.URLEncoder;
import java.util.*;
import java.rmi.RemoteException;
import java.lang.reflect.*;
import javax.servlet.http.*;

/**
 *  An abstract class that encapsulates the various paths of
 *  execution involved in fulfilling a servlet request. What is
 *  here is meant to be generic to the "typical" web app.
 *  Application-specific functionality will be supplied
 *  by concrete subclasses.
 *  @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */
abstract public class ServletInteraction implements NiggleConstants
{
    /*
     * constants associated with a ServletInteraction instance
     */
    public final HttpServletRequest request;

    /**
     * Our HTML response back to the user.
     */
    public final HttpServletResponse response;

    /**
     * The information that contains the niggle servlet config info.
     */
    public final NiggleConfig config;

    /**
     * numeric session id.
     */
    protected String sessionID;

    /**
     * Did we create a new session?
     * used for logging purposes.
     */
    protected boolean sessionCreated;

    /**
     * Login ID of the current user.
     */
    protected String userID;

    /**
     * a parameter from the servlet request used to hold supplemental information
     */
    protected String suppInfoParam;

    /**
      * locale of the server, ideally of the user
      */
    protected Locale locale = Locale.getDefault();

    /**
     * name of the current action being executed
     * by this ServletInteraction instance
     */
    protected String action;

    /**
     * The Page template object that we are working with.
     */
    protected Page page;

    /**
     * Indicates whether we handled the request by redirecting. A bit hacky.
     */
    protected boolean hasRedirected;

    public ServletInteraction(HttpServletRequest request,
                              HttpServletResponse response,
                              NiggleConfig config)
        throws IOException
    {
        this.request = request;
        this.response = response;
        this.config = config;
        this.sessionID = getParameter(SESSION_ID_KEY);
        this.userID = getParameter(USER_ID_KEY).toLowerCase().trim();
        this.action = getParameter(ACTION_KEY);
        if (!hasContent(action))
        {
            action = "default";
        }
        suppInfoParam = getParameter(SUPP_INFO_PARAM_KEY);
    }

    public final Locale getLocale()
    {
        return locale;
    }

    /**
     *  Message handler.  "Returns" a page that displays the
     *  message and title we pass in.  Provides a link back
     *  to the referring page.
     */
    private void displayMessage(String nextAction, String title, String message, boolean error)
    throws IOException
    {
        String template = error ? ERROR_TEMPLATE : MESSAGE_TEMPLATE;
        page = getPage(template);
        page.expose("title", title);
        page.expose("message", message);
        page.expose("returnURL", getURL(nextAction));
    }

    protected void displayMessage(String nextAction, String title, String message)
    throws IOException
    {
        displayMessage(nextAction, title, message, false);
    }

    protected void displayError(String nextAction, String title, String message)
    throws IOException
    {
        displayMessage(nextAction, title, message, true);
    }

    /**
     *  Throw up the login page template,
     */
    protected void displayLogin() throws IOException
    {
        page = getPage(LOGIN_TEMPLATE);

        // Set the following two boolean variables
        // to expose information about the reason for a login failure.
        boolean invalidSession = hasSession();
        boolean loginFailure = hasContent(userID) && !invalidSession;
        page.expose("invalidSession", invalidSession);
        page.expose("loginFailure", loginFailure);
    }

    /**
     * Builds complete URL with query string
     * corresponding to the given action and current context.
     * e.g. action=home&username=joe&sid=23213
     */
    public String getURL(String action)
    {
        if (!hasContent(action))
            action = "default";
        String queryString = "";
        String sidPart ="";
        String uidPart = "";
        // If this is the default action, we make it implicit
        if ( !action.equals("default"))
        {
            queryString += (ACTION_KEY + "=" + action + "&");
        }
        if (hasSession())
        {
            uidPart =  USER_ID_KEY + "=" + URLEncoder.encode(userID) + "&";
            sidPart = SESSION_ID_KEY + "=" + sessionID;
        }
        if (!action.equals("login"))
        {
            queryString = queryString + uidPart + sidPart;
        }
        String result = request.getRequestURI();
        char charSep = (result.indexOf("?") == -1) ? '?' : '&';
        if ( queryString.length()>0 )
            result = result + charSep + queryString;
        if ( result.endsWith("&") )
            result = result.substring(0, result.length() -1);
        return result;
    }

    /**
     * Check for presence of data field in CGI input.
     * wrapper for the ServletRequest's getParameter method
     * returns "" instead of null if the param is not found,
     * avoids annoying glitches.
     */
    public String getParameter(final String s)
    {
        if ( !hasContent(s) )
            return "";
        String result = request.getParameter(s);
        return (result != null) ? result : "";
    }


    /**
     * @return whether the param is present in the servlet request.
     */
    public boolean hasParameter(String param)
    {
        return request.getParameter(param) != null;
    }

    protected void invalidateSession() throws IOException
    {
        SessionManager sm = config.getSessionManager();
        if (sm != null)
            sm.invalidateSession(sessionID);
        userID = "";
        sessionID = "";
    }

    /**
     * check for empty strings.
     * @return true if string at least one char (possibly blank).
     */
    public static boolean hasContent(String s)
    {
        return s != null && s.length() > 0;
    }

    /**
     * Finally output the page as a result of a servlet interaction
     * We also touch the user's session if there is a sessionID.
     */
    protected void output() throws IOException
    {
        if ( !hasRedirected )
        {
            if ( page == null )
            {
                displayError("default", "ERROR", "Could not process your request");
            }
            exposeDefaultVariables();
            page.outputPage(response, action.equals("static"));
            if (hasSession())
            {
                SessionManager sm = config.getSessionManager();
                if (sm != null)
                    sm.touchSession(sessionID);
            }
        }
    }

    protected void createNewSession() throws IOException
    {
        SessionManager sm = config.getSessionManager();
        this.sessionID = sm.newSession(userID).getUniqueKey();
        this.sessionCreated = true;
    }

   /**
    * Method to call if a given action
    * requires a valid session.
    * NB: If there is no valid session info, but there is enough
    * info in the request to create a session
    * (typically id/password or possibly a cookie)
    * then create a new session. Otherwise, present a login page.
    *
    * @return true/false indicating whether the
    * login/session info is present. NB: If no session
    * manager has been initialized for this servlet,
    * this method just returns true.
    */
    protected boolean checkSessionInfo() throws IOException
    {
        SessionManager sm = config.getSessionManager();
        if (sm == null || sm.isValidSession(sessionID, userID))
           return true;
        if (hasValidLoginInfo())
        {
            createNewSession();
            return true;
        }
        displayLogin();
        return false;
    }

    /**
     *  A hook that can be overridden to make sure that
     * certain top-level template variables get exposed.
     */
    protected void exposeDefaultVariables() throws IOException
    {
        page.expose("loginURL", getURL("login"));
        page.expose("logoutURL", getURL("logout"));
    }

    /**
     * default handler for the "login" action.
     * Simply throws up a login page.
     */
    public void execLogin() throws IOException
    {
        displayLogin();
    }

    /**
     *  default logout handler.
     */
    public void execLogout() throws IOException
    {
        invalidateSession();
        page = getPage(LOGOUT_CONFIRM_TEMPLATE);
    }

    /**
     * A default handler for "action=static", which
     * will simply display a static page. Actually,
     * the page is <em>mostly</em> static, insofar as the
     * exposeDefaultVariables() method still gets called,
     * so the default template variables for the servlet
     * are still available to the people doing the GUI templates.
     */
    public void execStatic() throws IOException
    {
        String staticPage = suppInfoParam;
        if ( !hasContent(staticPage) )
            staticPage = "index";
        if ( !staticPage.endsWith(".html") )
            staticPage += ".html";
        try {
            page = getPage(staticPage);
        }
        catch (Exception e) {
            displayError(null,
                        "Navigation error",
                        "The page you requested is not available.");
            Logger.logException(e);
        }
    }

    /**
     * default handler for requests in which no
     * action is defined. This base implementation
     * assumes the "static" action.
     * see #execStatic()
     */
    public void execDefault() throws IOException
    {
        execStatic();
    }

    /**
     * @return the Page object for a given template.
     * delegates to the PageFactory object.
     */
    protected Page getPage(String template) throws IOException
    {
        Page result = config.getPageFactory().getPage(template, this);
        if (result != null)
           result.expose("template", template);
        return result;
    }

    /**
     * Indicates whether we have a session.
     */
    protected boolean hasSession()
    {
        return hasContent(sessionID);
    }

    /**
     * redirects the browser to the URL corresponding
     * to a given action.
     */
    protected void redirectAction(String action) throws IOException
    {
        response.sendRedirect(getURL(action));
        this.hasRedirected = true;
    }

    /**
     * A convenient method that exposes all of the
     * servlet request's parameters as template variables
     */
    protected void reexposeServletParams()
    {
        for ( Enumeration e = request.getParameterNames(); e.hasMoreElements(); )
        {
            String param = (String) e.nextElement();
            page.expose(param, request.getParameter(param));
        }
    }

    /**
     * Override this method to check whether the servlet
     * request has the info for a login (thus allowing us
     * to create a new session). In this base implementation,
     * this always returns false. Subclasses will likely
     * override this.
     */
    protected boolean hasValidLoginInfo()
    throws IOException
    {
        return false;
    }

    /**
     * Dispatches the request based on the action
     * This actually delegates to the NiggleConfig
     * object that does some reflective magic
     * to invoke the method this.execFoo() for action "foo".
     */
    protected void dispatch() throws IOException
    {
        try {
            config.dispatchAction(this);
        }
        catch (IOException e) {
            recover(e);
        }
    }

    /**
     * override this method to provide some default
     * exception handling for when an execXXX method
     * fails for whatever reason.
     * NB: This default implementation simply
     * rethrows the exception.
     */
    protected void recover(IOException e) throws IOException
    {
        throw e;
    }
}
