/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.templates.freemarkerimpl;

import com.niggle.data.Record;
import com.niggle.templates.*;
import com.niggle.util.*;
import freemarker.template.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;


/**
 * An implementation of the com.niggle.templates.Page interface
 * that uses the freemarker template library.
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */
public class FreemarkerPage implements Page
{
    protected LazyHashModel model;
    private Template pageTemplate;
    private String contentType = "text/html";

    /**
     *  Creates a new FreemarkerPage.
     *  @param pageTemplate the template to use.
     *  @param model the model to use.
     *
     *  @param type the type of data to output.  See class OutputTypes
     *  for a list of supported types.  May NOT be null.
     */
    public FreemarkerPage(Template pageTemplate, LazyHashModel model,
                          String contentType)
        throws IOException
    {
        this(pageTemplate, model);
        this.contentType = contentType;
    }

    public FreemarkerPage(Template pageTemplate, LazyHashModel model)
        throws IOException
    {
        Assert.assert(pageTemplate != null, "pageTemplate != null");
        this.pageTemplate = pageTemplate;
        if ( model == null )
            model = new LazyHashModel();
        model.setContentType(contentType);
        this.model = model;
    }

    /**
     *  Prepare the cooked HTML, filling in the parameters
     *  with data values, and send it to the end user
     *  via the servlet's response stream.
     *
     *  @param response handle to where the output should go.
     *  @exception IOException
     */
    public void outputPage(HttpServletResponse response, boolean isStatic)
        throws IOException
    {
        response.setContentType(contentType);
        if (!isStatic)
        {
            response.setHeader("Pragma", "no-cache"); // http 1.0
            response.setHeader("Cache-Control", "no-cache"); // http 1.1
        }
        PrintWriter out = response.getWriter();
        pageTemplate.process(model, out);
        out.close();
    }

    /**
     * @return the processed HTML as a String
     */
    public String toString()
    {
        StringWriter sw = new StringWriter();
        pageTemplate.process(model, new PrintWriter(sw));
        return sw.toString();
    }

    /**
     * expose internal data, publishing it to make it
     * available to a Freemarker Template.
     *
     * @param key    name of the field in the template
     * @param item   source of the data to populate the template
     */
    public void expose(String key, Object item)
    {
        if (item == null)
            return;
        if (item instanceof Record)
            model.put(key, (Record) item);
        else if (item instanceof TemplateModel)
            model.put(key, (TemplateModel) item);
        else if (item instanceof Boolean)
            model.put(key, ((Boolean) item).booleanValue());
        else if (item instanceof List)
            model.put(key, (List) item);
        else
            model.put(key, item.toString());
    }

    /**
     * Convenience method really, since all Freemarker
     * scalar variables are strings. Exposes the boolean
     * value as "true" or "". (Note that the empty string
     * in freemarker is essentially <em>false</em>
     *
     * @param key name of the field in the template
     * @param b sets to boolean value according to: b ? "true" : ""
     */
    public void expose(String key, boolean b)
    {
        model.put(key, b);
    }
}
