/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.templates.freemarkerimpl;

import freemarker.template.*;
import java.io.*;
import java.util.*;
import com.niggle.util.Util;
import javax.servlet.ServletContext;

/**
 * A version of the Freemarker TemplateCache that fishes out a
 * different template based on locale information.
 * (should be rolled into Freemarker actually)
 * @author <a href="mailto:jrevusky@bigfoot.com">Jonathan Revusky</a>
 */

public class LocalizedFileTemplateCache extends FileTemplateCache
implements LocalizedTemplateCache
{
    public Template getTemplate(String filename, Locale locale)
    {
        String[] fnames = getFilenames(filename, locale);
        Template result = lookInContext(fnames);
        if (result != null)
        {
            result.setTemplateCache(this);
        }
        return result;
    }

    /**
     * Run through the list of filenames, which
     * are assumed to be listed in order of more
     * specifically qualified down to default,
     * and try to get the resource relative to
     * the CLASSPATH.
     * @return null on failure.
     */
    private Template lookInContext(String[] fnames)
    {
        for (int i = 0; i < fnames.length; i++)
        {
            if (fnames[i] == null)
                continue;
            if (explicitDirectory != null)
            {
                Template tm = getTemplate(fnames[i]);
                if (tm != null)
                   return tm;
            }
            String filename = fnames[i];
            if (Util.hasContent(path))
            {
                filename = path + "/" + filename;
            }
            if (ourCache.containsKey(filename))
            {
                Template tm = (Template) ourCache.get(filename);
                if (tm == null)
                   continue;
                return tm;
            }
            InputStream is = Util.getInputStream(filename, cl);
            if (is == null)
            {
                // Okay, stick a null in so we won't waste
                // our time next time!.
                ourCache.put(filename, null);
            }
            else try {
                Template tm = new Template(is);
                ourCache.put(filename, tm);
                return tm;
            }
            catch (Exception e) {
                ourCache.put(filename, null);
            }
        }
        return null; // Couldn't find anything!
    }

    public void setPath(String path)
    {
        this.path = path;
        File dir = new File(path);
        char firstChar = 'a';
        if (path.length() >0) 
            firstChar = path.charAt(0);
        boolean absolute = dir.isAbsolute() || firstChar=='/' || firstChar=='\\';
        if (absolute && dir.canRead())
        {
            this.explicitDirectory = dir;
            super.setDirectory(dir);
            startAutoUpdate();
        }
    }

    private String[] getFilenames(String filename, Locale locale)
    {
        String extension = "";
        String basename = filename;
        String variant = locale.getVariant();
        String country = locale.getCountry();
        String lang = locale.getLanguage();
        Template result = null;
        int dotIndex = filename.lastIndexOf('.');
        if (dotIndex > 0)
        {
            basename = filename.substring(0, dotIndex);
            extension = filename.substring(dotIndex);
        }
        //Level of full specification.
        // We try them in order.
        String filename1 = Util.hasContent(variant) ?
                              basename + "_"
                              + lang + "_"
                              + country + "_"
                              + variant + extension
                              : null;
        String filename2 = Util.hasContent(country) ?
                              basename + "_"
                              + lang + "_"
                              + country + extension
                              : null;
        String filename3 = Util.hasContent(lang) ?
                              basename + "_"
                              + lang + extension
                              : null;
        return new String[] {
            filename1, filename2, filename3, filename
        };
    }

    void setClassLoader(ClassLoader cl)
    {
        this.cl = cl;
    }


    private final HashMap ourCache = new HashMap();
    private String path ="";
    private File explicitDirectory;
    private ClassLoader cl;
}
