/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.niggle.util;

import java.io.*;
import java.util.*;
import java.text.*;

/**
 *  A file lock.  Since proper file locking is problematic (PhD thesis goes
 *  here), we use this class throughout the MediMOM systems to manage
 *  access to shared flat files.  This mechanism is used
 *  <b>by convention</b>.
 *  <p>
 *  This class will attempt to acquire a lock for at most MAX_LOCK_LIFE
 *  seconds.  If that fails, an error is written to the log file, and the
 *  lock is stolen.  This implies that we never expect any of our processes
 *  to spend more than MAX_LOCK_LIFE writing to a file.  If a lock gets
 *  to be that old, we assume the process or thread that created it died
 *  unexpectedly.
 *  <p>
 *  MAX_LOCK_LIFE should be set to a value that we would consider to be
 *  beyond the maximum acceptable time for any of our write operations.
 *  <p>
 *  <b>
 *  NB: This locking system is a perfect semaphore mechanism within
 *  a single process.  It is pretty good between processes, but not
 *  perfect.  The only way it will be perfect is if the locking is
 *  managed by a single process.
 *  </b>
 * @author Paul Murphy murphy@niggle.com
 */

public class FileLock
{
    /*
     *  CONSTANTS.
     */

    private static final int DEFAULT_MAX_LOCK_LIFE = 20;    /* in secs. */
    private static final int DEFAULT_LOCK_CHECK_WAIT = 20;  /* in millisecs. */

    /*
     *  CLASS VARS.
     */

    private static int _maxLockLife = DEFAULT_MAX_LOCK_LIFE;
    private static int _lockCheckWait = DEFAULT_LOCK_CHECK_WAIT;

    /*
     *  IVARS.
     */

    private String _lockFileName;
    private String _lockTimeStamp;

    /**
     *  Illegal constructor.
     */
    private FileLock()
    {
    }

    /**
     *  Designated constructor.
     *
     *  @arg filename name of the file for which the lock is being
     *  created.
     *  <br><b>require:</b> <code>(filename != null) &&
     *  (filename.length() > 0)</code>
     *
     *  @return a new instance, or <code>null</code>
     *  <br><b>ensure:</b> <code>(result != null) =>
     *  (_lockFileName != null)</code>
     *
     *  @exception FileLockException thrown when a lock cannot be acquired.
     */
    public FileLock(String filename) throws FileLockException
    {
        super();
    /* Set ivar. */
        _lockFileName = filename + ".lck";

        String oldTS = null;
        String err = null;
        File lockFile = new File(_lockFileName);

    /*
     *  If the lock file exists, wait till it disappears.
     */
        if (lockFile.exists() == true)
        {
            long t = System.currentTimeMillis();

            while ((t + (_maxLockLife * 1000)) > (System.currentTimeMillis()))
            {
                Logger.log("Waiting for lock for " + filename + ".",
                           Logger.LOG_WARN);
                try
                {
                    Thread.sleep(500);
                }
                catch (InterruptedException e)
                {
                    Logger.log("InterruptedException caught.",
                               Logger.LOG_INFO);
                }
                if (lockFile.exists() != true)
                    break;
            }
        }

    /*
     *  Generate the time stamp we're going to use.
     */
        _lockTimeStamp = DateFormat.getDateTimeInstance().format(new Date());

    /*
     *  If it's not gone, we steal it.
     */

        if (lockFile.exists() == true)
        {
            oldTS = null;
            try
            {
                BufferedReader br =
                new BufferedReader(new FileReader(lockFile));
                oldTS = br.readLine();
                br.close();
            }
            catch (FileNotFoundException e)
            {
                Logger.log("Lock for " + filename + " disappeared (1).",
                           Logger.LOG_ERROR);
            }
            catch (IOException e)
            {
                Logger.log("Lock for " + filename + " disappeared (2).",
                           Logger.LOG_ERROR);
            }

            Logger.log("Stealing lock (" + oldTS + ") at "
                       + _lockTimeStamp + ".", Logger.LOG_WARN);
        }

    /*
     *  Write a timestamp into the file.  If this fails in any way,
     *  we BAIL.
     */
        try
        {
            BufferedWriter bw =
            new BufferedWriter(new FileWriter(lockFile));
            bw.write(_lockTimeStamp + "\n");
            bw.close();
        }
        catch (IOException e)
        {
            err = "Unable to create lock for " + filename + ".";
            Logger.log(err, Logger.LOG_ERROR);
            throw new FileLockException(err);
        }

    /*
     *  This is an extra security blanket.  It's a bit cheesy but
     *  buys us a bit of safety until we have a more robust solution.
     *  We sleep for _lockCheckWait ms and then read the timestamp from
     *  the lock file we just created.  If it's different, we assume
     *  we've been stomped on by another process & raise an
     *  exception.
     */
        try
        {
            Thread.sleep(_lockCheckWait);
        }
        catch (InterruptedException e)
        {
            Logger.log("InterruptedException caught.", Logger.LOG_INFO);
        }

        try
        {
            BufferedReader br =
            new BufferedReader(new FileReader(lockFile));
            oldTS = br.readLine();
            br.close();
        }
        catch (FileNotFoundException e)
        {
            err = "Lock for " + filename + " was stomped (1).";
            Logger.log(err, Logger.LOG_ERROR);
            throw new FileLockException(err);
        }
        catch (IOException e)
        {
            err = "Lock for " + filename + " was stomped (2).";
            Logger.log(err, Logger.LOG_ERROR);
            throw new FileLockException(err);
        }
        if (!_lockTimeStamp.equals(oldTS))
        {
            err = "Lock for " + filename + " was stomped (3).";
            Logger.log(err, Logger.LOG_ERROR);
            throw new FileLockException(err);
        }
    }

    /**
     *  Deletes the file associated with the current lock object.  If the
     *  file does not exist, or contains an invalid timestamp, logs an
     *  error.
     */
    public void dispose()
    {
        String oldTS = null;
        File lockFile = new File(_lockFileName);

        if (lockFile.exists() == false)
        {
            Logger.log("Lock " + _lockFileName + " disappeared. (1)",
                       Logger.LOG_ERROR);
        }
        else // (lockFile.exist() == true)
        {
        /*
         *  Read the contents of the lock file.
         */
            try
            {
                BufferedReader br =
                new BufferedReader(new FileReader(lockFile));
                oldTS = br.readLine();
                br.close();
            }
            catch (FileNotFoundException e)
            {
                Logger.log("Lock " + _lockFileName + " disappeared. (1)",
                           Logger.LOG_ERROR);
            }
            catch (IOException e)
            {
                Logger.log("Problem reading " + _lockFileName + ".",
                           Logger.LOG_ERROR);
            }

        /*
         *  If they aren't what we expect, report the problem.
         */
            if (!_lockTimeStamp.equals(oldTS))
            {
                Logger.log("Lock " + _lockTimeStamp + " (" + oldTS +
                           ") not what we expected: " + _lockTimeStamp + ".",
                           Logger.LOG_ERROR);
            }

        /*
         *  Delete the lock.
         */
            if (lockFile.delete() != true)
            {
                Logger.log("Unable to delete " + _lockFileName + ".",
                           Logger.LOG_ERROR);
            }
        }
    }

    /*
     *  CLASS METHODS.
     */

    /**
     *  Sets <code>_maxLockLife</code> to <code>newLife</code>, in seconds.
     *
     *  @param newLife the amount of time a lock can live.
     *  <br><b>require:</b> <code>newLife >= 1</code>
     */
    private static void setMaxLockLife(int newLife)
    {
        _maxLockLife = newLife;
    }

    /**
     *  Sets <code>_lockCheckWait</code> to <code>newWait</code>, in
     *  milliseconds.
     *
     *  @param newWait the amount of time to wait to make sure no other
     *  process stomped our lock.  In a well designed system, this
     *  value should be 0.
     *  <br><b>require:</b> <code>newLife >= 0</code>
     */
    private static void setLockCheckWait(int newWait)
    {
        _lockCheckWait = newWait;
    }
}



